/* ***********************************************

AXPY

memo:
Actually v2&v4 kernels are not needed.
v1 kernel with AT is enought to achieve best throughput.

  *********************************************** */   

#include "mublas_devel.h"
#include "mublas_axpy_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   
__host__ int32_t
MUBLAS_AXPY (
	mublasHandle_t* mh,	
	const SIZE_T n,
	const TYPE0 *alpha,
	const TYPE0 *x,
	const SIZE_T incx,
	TYPE0 *y,
	const SIZE_T incy
) {
	int32_t kernel = 0;
	dim3 threads;
	dim3 grid;

	// ---------------------------------------------
	// 1. Shortcut
    if (n == 0 || *alpha == ZERO0) 
		return 0;
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration Phase 0
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if (
		(((long)x%(128/sizeof(TYPE0)))%16 == 0) &&
		(((long)y%(128/sizeof(TYPE0)))%16 == 0) 
		) {
		if (VEC4 && n%4 == 0)
			mh->N_T = 4;
		else if (VEC2 && n%2 == 0)
			mh->N_T = 2;
	}
	mh->n = n;
	// ---------------------------------------------

	// debug mode
	if (mh->kernelnumber > 8000) {
		mh->N_T = 1;
		if (mh->kernelnumber == 8120) mh->N_T = 2;
		if (mh->kernelnumber == 8140) mh->N_T = 4;
	} 

	// ---------------------------------------------
	// 3. Kernel Configuration Phase 1
	kernel = 999; // default=CUBLAS;
	if (incx == 1 && incy == 1) {
		mh->BlkDim = 1;
		mh->func_getShmPerBlk = &getShmPerBlk_AXPY;
		mh->func_getBlkPerGrd = &getBlkPerGrd_AXPY;
		#include "mublas_axpy.incl"
		if (mh->N_T == 4)
			kernel = 140;
		else if (mh->N_T == 2)
			kernel = 120;
		else
			kernel = 110;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 4. Kernel Configuration Phase 2
	switch (kernel) {
		case 140:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_AXPY_DEVICE_v4);
			cudaFuncSetCacheConfig (MUBLAS_AXPY_DEVICE_v4, cudaFuncCachePreferL1);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx);
			grid = dim3 (mh->BlkPerDev);
			MUBLAS_AXPY_DEVICE_v4 <<< grid, threads >>> (n, *alpha, x, y);
			break;
		case 120:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_AXPY_DEVICE_v2);
			cudaFuncSetCacheConfig (MUBLAS_AXPY_DEVICE_v2, cudaFuncCachePreferL1);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx);
			grid = dim3 (mh->BlkPerDev);
			MUBLAS_AXPY_DEVICE_v2 <<< grid, threads >>> (n, *alpha, x, y);
			break;
		case 110:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_AXPY_DEVICE_s1);
			cudaFuncSetCacheConfig (MUBLAS_AXPY_DEVICE_s1, cudaFuncCachePreferL1);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx);
			grid = dim3 (mh->BlkPerDev);
			MUBLAS_AXPY_DEVICE_s1 <<< grid, threads >>> (n, *alpha, x, y);
			break;
		default:
			#if defined (CUBLAS_AXPY)
			cublasHandle_t ch;
			cublasCreate (&ch);
			cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
			CUBLAS_AXPY (ch, n, alpha, x, incx, y, incy);
			#else
			printf ("mublas error: not implemented");
			#endif
			break;
	}
	// ---------------------------------------------
		
	if (mh->kernelnumber < 8000) mh->kernelnumber = kernel;

	return 0;
}

