/* ***********************************************

AXPY

memo:
Actually v2&v4 kernels are not needed.
v1 kernel with AT is enought to achieve best throughput.

  *********************************************** */   

#include "mublas_axpy.def"

/* ***********************************************
    Macro
   *********************************************** */   
#if USE_NC_X_AXPY
#define RESTRICT_X __restrict__
#define LOAD1_X LOAD1_NC
#define LOAD2_X LOAD2_NC
#define LOAD4_X LOAD4_NC
#else
#define RESTRICT_X
#define LOAD1_X LOAD1___
#define LOAD2_X LOAD2___
#define LOAD4_X LOAD4___
#endif

#define LOAD1_Y LOAD1___
#define LOAD2_Y LOAD2___
#define LOAD4_Y LOAD4___

/* ***********************************************
	Functions for determining Tx 
   *********************************************** */   
__host__ __inline__ void getBlkPerGrd_AXPY ( mublasHandle_t* mh) {
	mh->BlkPerDev = (SIZE_T) ceil ((float)mh->n / (float)(mh->N_T * mh->Tx));
}

__host__ __inline__ void getShmPerBlk_AXPY ( mublasHandle_t* mh) {
	mh->ShmPerBlk = 0.;
}

/* ***********************************************
	kernel=v4
   *********************************************** */   

__global__ void MUBLAS_AXPY_DEVICE_v4 (
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_X x,
	TYPE0 *y
) {
#if VEC4
	const SIZE_T N_T = 4;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T rowid = blockIdx.x * Tx * N_T + iTx * N_T;
	TYPE4 Xr4, Yr4;

	if (rowid < n) {
		LOAD4_X (Xr4, &x[rowid]);
		LOAD4_Y (Yr4, &y[rowid]);
		Yr4 = ADD4 (Yr4, MUL14 (alpha, Xr4));
		(TYPE4&) y[rowid] = Yr4;
	}
#endif
}

/* ***********************************************
	kernel=v2
   *********************************************** */   

__global__ void MUBLAS_AXPY_DEVICE_v2 (
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_X x,
	TYPE0 *y
) {
#if VEC2
	const SIZE_T N_T = 2;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T rowid = blockIdx.x * Tx * N_T + iTx * N_T;
	TYPE2 Xr2, Yr2;

	if (rowid < n) {
		LOAD2_X (Xr2, &x[rowid]);
		LOAD2_Y (Yr2, &y[rowid]);
		Yr2 = ADD2 (Yr2, MUL12 (alpha, Xr2));
		(TYPE2&) y[rowid] = Yr2;
	}
#endif
}


/* ***********************************************
	kernel=s1
   *********************************************** */   

__global__ void MUBLAS_AXPY_DEVICE_s1 (
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_X x,
	TYPE0 *y
) {
#if VEC1
	const SIZE_T Tx = blockDim.x;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T rowid = blockIdx.x * Tx + iTx;
	TYPE1 Xr1, Yr1;

	if (rowid < n) {
		LOAD1_X (Xr1, &x[rowid]);
		LOAD1_Y (Yr1, &y[rowid]);
		y[rowid] = ADD1 (Yr1, MUL1 (alpha, Xr1));
	}

#endif
}

#undef RESTRICT_X
#undef LOAD1_X
#undef LOAD2_X
#undef LOAD4_X
#undef LOAD1_Y
#undef LOAD2_Y
#undef LOAD4_Y

