/* ***********************************************
 
	TRMV for CUDA 

   *********************************************** */   

#include "mublas_devel.h"
#include "mublas_trmv_lnn_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   

__host__ int32_t
MUBLAS_TRMV (
	mublasHandle_t* mh,	
	const char uplo,
	const char trans,
	const char diag,
	const SIZE_T n,
	const TYPE0 *a,
	const SIZE_T lda,
	TYPE0 *x,
	const SIZE_T incx
) {
	SIZE_T kernel = 0;
	dim3 threads;
	dim3 grid;

	// ---------------------------------------------
	// 1. Shortcut
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration Phase 0
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if (
		(((long)a%(128/sizeof(TYPE0)))%16 == 0) &&
		(lda%mh->N_T == 0)
		) {
		if (VEC4 && n%4 == 0)
			mh->N_T = 4;
		else if (VEC2 && n%2 == 0)
			mh->N_T = 2;
	}
	mh->n = n;
	// ---------------------------------------------

	// ---------------------------------------------
	// 3. Kernel Configuration Phase 1
	kernel = 999; // default=CUBLAS;
	if (incx == 1) {
		if (uplo == 'L' || uplo == 'l') {
			if (trans == 'N' || trans == 'n') {
				if (diag == 'N' || diag == 'n') {
					mh->BlkDim = 2;
					mh->func_getShmPerBlk = &getShmPerBlk_TRMV_LNN;
					mh->func_getBlkPerGrd = &getBlkPerGrd_TRMV_LNN;
					#include "mublas_trmv_lnn.incl"
					if (mh->N_T == 4) 
						kernel = 140;
					else if (mh->N_T == 2)
						kernel = 120;
					else
						kernel = 110;
				}
			}
		}
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 4. Kernel Configuration Phase 2
	switch (kernel) {
		// LNN-kernel
		case 140:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_TRMV_LNN_DEVICE_v4);
			cudaFuncSetCacheConfig (MUBLAS_TRMV_LNN_DEVICE_v4, cudaFuncCachePreferShared);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->BlkPerDev);
			MUBLAS_TRMV_LNN_DEVICE_v4 <<< grid, threads, mh->ShmPerBlk >>> (n, a, lda, x, (TYPE0*)mh->work_addr, (TYPE0*)mh->work_addr2);
			threads = dim3 (64);
			grid = dim3 ((SIZE_T) ceil ((float)mh->n / 64.));
			MUBLAS_TRMV_LNN_DEVICE_final <<< grid, threads >>> (n, (TYPE0*)mh->work_addr, (TYPE0*)mh->work_addr2, x);
			break;
		case 120:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_TRMV_LNN_DEVICE_v2);
			cudaFuncSetCacheConfig (MUBLAS_TRMV_LNN_DEVICE_v2, cudaFuncCachePreferShared);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->BlkPerDev);
			MUBLAS_TRMV_LNN_DEVICE_v2 <<< grid, threads, mh->ShmPerBlk >>> (n, a, lda, x, (TYPE0*)mh->work_addr, (TYPE0*)mh->work_addr2);
			threads = dim3 (64);
			grid = dim3 ((SIZE_T) ceil ((float)mh->n / 64.));
			MUBLAS_TRMV_LNN_DEVICE_final <<< grid, threads >>> (n, (TYPE0*)mh->work_addr, (TYPE0*)mh->work_addr2, x);
			break;
		case 110:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_TRMV_LNN_DEVICE_s1);
			cudaFuncSetCacheConfig (MUBLAS_TRMV_LNN_DEVICE_s1, cudaFuncCachePreferShared);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->BlkPerDev);
			MUBLAS_TRMV_LNN_DEVICE_s1 <<< grid, threads, mh->ShmPerBlk >>> (n, a, lda, x, (TYPE0*)mh->work_addr, (TYPE0*)mh->work_addr2);
			threads = dim3 (64);
			grid = dim3 ((SIZE_T) ceil ((float)mh->n / 64.));
			MUBLAS_TRMV_LNN_DEVICE_final <<< grid, threads >>> (n, (TYPE0*)mh->work_addr, (TYPE0*)mh->work_addr2, x);
			break;
		// CUBLAS
		default:
			#if defined (CUBLAS_TRMV)
			cublasHandle_t ch;
			cublasCreate (&ch);
			cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
			cublasFillMode_t uploAcublas = CUBLAS_FILL_MODE_UPPER;
			cublasOperation_t tranAcublas = CUBLAS_OP_N;
			cublasDiagType_t diagAcublas = CUBLAS_DIAG_NON_UNIT;
			if (uplo == 'L' || uplo == 'l') uploAcublas = CUBLAS_FILL_MODE_LOWER;
			if (trans == 'T' || trans == 't') tranAcublas = CUBLAS_OP_T;
			if (diag == 'U' || trans == 'u') diagAcublas = CUBLAS_DIAG_UNIT;
			CUBLAS_TRMV (ch, uploAcublas, tranAcublas, diagAcublas, n, a, lda, x, incx);
			#else
			printf ("mublas error: not implemented");
			#endif
			break;
	}
	// ---------------------------------------------

	if (mh->kernelnumber < 8000) mh->kernelnumber = kernel;
	return 0;
}

