#ifndef PREC_S_H
#define PREC_S_H

#ifdef PREC_S
	// host
	#define MUBLAS_CSRMV mublasScsrmv
	#define MUBLAS_GEMM mublasSgemm
	#define MUBLAS_GEMV mublasSgemv
	#define MUBLAS_TRMV mublasStrmv
	#define MUBLAS_AXPY mublasSaxpy
	#define MUBLAS_DOT  mublasSdot
	#define MUBLAS_SCAL mublasSscal
	#define MUBLAS_COPY mublasScopy
	#define CUSPARSE_CSRMV cusparseScsrmv
	#define CUBLAS_GEMM cublasSgemm
	#define CUBLAS_GEMV cublasSgemv
	#define CUBLAS_TRMV cublasStrmv
	#define CUBLAS_AXPY cublasSaxpy
	#define CUBLAS_DOT  cublasSdot
	#define CUBLAS_SCAL cublasSscal
	#define CUBLAS_COPY cublasScopy
	#define TYPE0 float
	#define ZERO0 0.
	#define ONE0 1.

	// vec1
	#define VEC1 1
	#define TYPE1 TYPE0
	#define ZERO1 ZERO0
	#define ADD1(A,B) ((A)+(B))
	#define MUL1(A,B) ((A)*(B))
	#define MAD1(A,B,C) (((A)*(B))+(C))

	#define TYPEIN1 TYPE1
	#define ZEROIN1 ZERO1
	#define LOAD1___ ld_global_f32
	#define LOAD1_NC ld_global_nc_f32

	// vec2
	#define VEC2 1
	#define TYPE2 float2
	#define ZERO2 make_float2(0.,0.)
	#define ADD2 cuSadd2
	#define MUL12 cuSmul12
	#define MAD122 cuSmad122
	#define MAD212 cuSmad212
	#define MAD2 cuSmad2

	#define TYPEIN2 TYPE2
	#define ZEROIN2 ZERO2
	#define LOAD2___ ld_global_v2_f32
	#define LOAD2_NC ld_global_nc_v2_f32

	// vec4
	#define VEC4 1
	#define TYPE4 float4
	#define ZERO4 make_float4(0.,0.,0.,0.)
	#define ADD4 cuSadd4
	#define MUL14 cuSmul14
	#define MAD144 cuSmad144
	#define MAD414 cuSmad414
	#define MAD4 cuSmad4

	#define TYPEIN4 TYPE4
	#define ZEROIN4 ZERO4
	#define LOAD4___ ld_global_v4_f32
	#define LOAD4_NC ld_global_nc_v4_f32

#endif

// === Shuffle =====================================================
__device__ __forceinline__ void shuffleAdd (float &val, const int n) {
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	val += __shfl_xor(val, n, WrpSize);
#endif
}
// =================================================================

// === Load & Store ================================================
__device__ __forceinline__ void ld_global_v4_f32 (float4 &reg, const float* addr) {
#if USE_INLINE_ASM
	asm ("ld.global.v4.f32 {%0,%1,%2,%3}, [%4+0];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
#else
	reg = (float4&)*addr;
#endif
}

__device__ __forceinline__ void ld_global_v2_f32 (float2 &reg, const float* addr) {
#if USE_INLINE_ASM
	asm ("ld.global.v2.f32 {%0,%1}, [%2+0];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
#else
	reg = (float2&)*addr;
#endif
}

__device__ __forceinline__ void ld_global_f32 (float &reg, const float* addr) {
#if USE_INLINE_ASM
	asm ("ld.global.f32 %0, [%1];": "=f"(reg): "l"(addr));
#else
	reg = *addr;
#endif
}

__device__ __forceinline__ void ld_global_nc_v4_f32 (float4 &reg, const float* addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v4.f32 {%0,%1,%2,%3}, [%4+0];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
	#else
	asm ("ld.global.v4.f32 {%0,%1,%2,%3}, [%4+0];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
	#endif
#else
	reg = (float4&)*addr;
#endif
}

__device__ __forceinline__ void ld_global_nc_v2_f32 (float2 &reg, const float* addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v2.f32 {%0,%1}, [%2+0];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
	#else
	asm ("ld.global.v2.f32 {%0,%1}, [%2+0];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
	#endif
#else
	reg = (float2&)*addr;
#endif
}

__device__ __forceinline__ void ld_global_nc_f32 (float &reg, const float* addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.f32 %0, [%1];": "=f"(reg): "l"(addr));
	#else
	asm ("ld.global.f32 %0, [%1];": "=f"(reg): "l"(addr));
	#endif
#else
	reg = *addr;
#endif
}
// =================================================================

// =================================================================

__device__ __forceinline__ float2
cuSadd2
	(const float2 a, const float2 b)
{
	return make_float2 (a.x+b.x, a.y+b.y);
}

__device__ __forceinline__ float2
cuSmul12
	(const float a, const float2 b)
{
	return make_float2 (a*b.x, a*b.y);
}

__device__ __forceinline__ float2
cuSmad122
	(const float a, const float2 b, const float2 c)
{
	return make_float2 (a*b.x+c.x, a*b.y+c.y);
}

__device__ __forceinline__ float2
cuSmad212
	(const float2 a, const float b, const float2 c)
{
	return make_float2 (a.x*b+c.x, a.y*b+c.y);
}

__device__ __forceinline__ float2
cuSmad2
	(const float2 a, const float2 b, const float2 c)
{
	return make_float2 (a.x*b.x+c.x, a.y*b.y+c.y);
}

__device__ __forceinline__ float4
cuSadd4
	(const float4 a, const float4 b)
{
	return make_float4 (a.x+b.x, a.y+b.y, a.z+b.z, a.w+b.w);
}

__device__ __forceinline__ float4
cuSmul14
	(const float a, const float4 b)
{
	return make_float4 (a*b.x, a*b.y, a*b.z, a*b.w);
}

__device__ __forceinline__ float4
cuSmad144
	(const float a, const float4 b, const float4 c)
{
	return make_float4 (a*b.x+c.x, a*b.y+c.y, a*b.z+c.z, a*b.w+c.w);
}

__device__ __forceinline__ float4
cuSmad414
	(const float4 a, const float b, const float4 c)
{
	return make_float4 (a.x*b+c.x, a.y*b+c.y, a.z*b+c.z, a.w*b+c.w);
}

__device__ __forceinline__ float4
cuSmad4
	(const float4 a, const float4 b, const float4 c)
{
	return make_float4 (a.x*b.x+c.x, a.y*b.y+c.y, a.z*b.z+c.z, a.w*b.w+c.w);
}

#endif
