#include "testing_common.h"
#include "testing_level2.h"
#include "testing_common.cpp"

int32_t
main (int32_t argc, char **argv)
{
	size_t i, iter = MINLOOP;
	size_t dim_mm, dim_nn, lda, lda_cpu;
	float sec;
	GPU_TYPE alpha, beta;
	GPU_TYPE *hst_A, *hst_X, *hst_Y;
	GPU_TYPE *hst_Y_t;
	GPU_TYPE *dev_A, *dev_X, *dev_Y;

	REFERENCE_TYPE *hst_A_r = 0, *hst_X_r = 0, *hst_Y_r = 0, alpha_r = REFERENCE_ZERO, beta_r = REFERENCE_ZERO;
	size_t size_r = sizeof (REFERENCE_TYPE);

// testing setup ------------------------------
	testingHandle_t th;
	testingCreate (argc, argv, &th);
	dim_initialize (&th);
	print_info1 (&th);
// --------------------------------------------

// library setup ------------------------------
	cudaEvent_t t_start, t_stop;
	cudaEventCreate (&t_start);
	cudaEventCreate (&t_stop);
	cudaError_t err;

	#if defined (USE_NVML)
	uint32_t temp1 = 0, temp2 = 0;
	nvmlDevice_t dev;
	nvmlInit();
	nvmlDeviceGetHandleByIndex(th.devno, &dev);
	do {
		sleep(1);
		nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp1);
	} while (temp1 >= TEMPLIMIT);
	#endif

	#if defined (CUBLAS) || (defined (CHECK_CUBLAS))
	cublasHandle_t ch;
	cublasCreate (&ch);
	cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
	#endif
	#if defined (MUBLAS)
	mublasHandle_t mh;
	mublasCreate (&mh);
	set_mublas_handler (&th, &mh);
	#elif defined (CUBLAS)
	if (th.cublasatomic == 1) 
		cublasSetAtomicsMode(ch, CUBLAS_ATOMICS_ALLOWED);
	else
		cublasSetAtomicsMode(ch, CUBLAS_ATOMICS_NOT_ALLOWED);
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	magma_init();
	#elif defined (CULA)
    culaInitialize();
	#elif defined (ASPEN)
	ASPEN_init(0);
	#else
	#endif
#if DEBUG_ALLSKIP == 0
// --------------------------------------------

// memory setup -------------------------------
	cudaMallocHost ((void **) &hst_A, th.word_size * th.dim_m * th.dim_n);
	cudaMallocHost ((void **) &hst_X, th.word_size * th.dim_n * th.incx);
	cudaMallocHost ((void **) &hst_Y, th.word_size * th.dim_m * th.incy);
	init_vector (1, &alpha);
	init_vector (1, &beta);
	init_vector (th.dim_m * th.dim_n, hst_A);
// Hermitian (for MAGMA)
	for (i=0; i<th.dim_m; i++) {
#if defined (PREC_C)
		hst_A[i*th.dim_m+i] = make_cuFloatComplex(hst_A[i*th.dim_m+i].x,0);
#elif defined (PREC_Z)
		hst_A[i*th.dim_m+i] = make_cuDoubleComplex(hst_A[i*th.dim_m+i].x,0);
#endif
	}
	init_vector (th.dim_n * th.incx, hst_X);
	init_vector (th.dim_m * th.incy, hst_Y);
	if (th.mode == 'c') {
		cudaMallocHost ((void **) &hst_Y_t, th.word_size * th.dim_m * th.incy);
		hst_A_r = (REFERENCE_TYPE *) malloc (size_r * th.dim_m * th.dim_n);
		hst_X_r = (REFERENCE_TYPE *) malloc (size_r * th.dim_n * th.incx);
		hst_Y_r = (REFERENCE_TYPE *) malloc (size_r * th.dim_m * th.incy);
		copy_vector (th.dim_m * th.dim_n, hst_A, hst_A_r);
		copy_vector (th.dim_n * th.incx, hst_X, hst_X_r);
		copy_vector (th.dim_m * th.incy, hst_Y, hst_Y_r);
		copy_vector (1, &alpha, &alpha_r);
		copy_vector (1, &beta, &beta_r);
	}
// --------------------------------------------

	print_info2 (&th);

// evaluation ---------------------------------
	dim_start (&th);
	while (1) {
		if ((th.const_m == 0 && th.dim_m > th.dim_stop) ||
		    (th.const_n == 0 && th.dim_n > th.dim_stop) ||
		    (th.const_k == 0 && th.dim_k > th.dim_stop)) break;

		if (th.tranA == 'T' || th.tranA == 't' || th.tranA == 'C' || th.tranA == 'c') {
			lda = th.dim_n;
			lda_cpu = th.dim_n;
			dim_mm = th.dim_n;
			dim_nn = th.dim_m;
		} else {
			lda = th.dim_m;
			if (th.const_m == 1) lda_cpu = th.dim_m;
			else lda_cpu = th.dim_stop;
			dim_mm = th.dim_m;
			dim_nn = th.dim_n;
		}

		#if PITCHALLOC
		size_t pitch;
		cudaMallocPitch ((void **) &dev_A, &pitch, th.word_size * lda, dim_nn);
		lda = pitch/th.word_size;
		#else
		cudaMalloc ((void **) &dev_A, th.word_size * dim_mm * dim_nn);
		#endif
		cudaMalloc ((void **) &dev_X, th.word_size * th.dim_n * th.incx);
		cudaMalloc ((void **) &dev_Y, th.word_size * th.dim_m * th.incy);

		cublasSetMatrix (dim_mm, dim_nn, th.word_size, hst_A, lda_cpu, dev_A, lda);
		cublasSetVector (th.dim_n * th.incx, th.word_size, hst_X, 1, dev_X, 1);
		cublasSetVector (th.dim_m * th.incy, th.word_size, hst_Y, 1, dev_Y, 1);

		#if defined (USE_NVML)
		nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp1);
		#endif

		#if defined (DIM1)
		printf ("--\t%zd\t--", (size_t)th.dim_n);
		#elif defined (DIM2)
		printf ("%zd\t%zd\t--", (size_t)th.dim_m, (size_t)th.dim_n);
		#endif
		get_routine_theoretial_performance (&th);

		// execution ---------------------------------
		if (th.mode == 'p') {
			#if PERF_ACCURATE
			iter = get_num_iter (&th);
			ROUTINE_TARGET(th,dim_mm,dim_nn,alpha,dev_A,lda,dev_X,beta,dev_Y); // dummy
			// +++++ measurement +++++++
			cudaEventRecord (t_start, 0);
			for (i = 0; i < iter; i++) ROUTINE_TARGET(th,dim_mm,dim_nn,alpha,dev_A,lda,dev_X,beta,dev_Y);
			cudaEventRecord (t_stop, 0);
			// +++++++++++++++++++++++++
			cudaEventSynchronize (t_stop);
			cudaEventElapsedTime (&sec, t_start, t_stop);
			sec = 1.0e-3 * sec / (float)iter;
			#else
			float sec_all_exec[MINLOOP];
			iter = MINLOOP;
			for (i = 0; i < iter; i++) {
				cudaEventRecord (t_start, 0);
				ROUTINE_TARGET(th,dim_mm,dim_nn,alpha,dev_A,lda,dev_X,beta,dev_Y);
				cudaEventRecord (t_stop, 0);
				cudaEventSynchronize (t_stop);
				cudaEventElapsedTime (&sec, t_start, t_stop);
				sec_all_exec[i] = sec;
			}
			sec = 0.;
			for (i = 0; i < iter; i++) {
				if (sec_all_exec[i] > sec)
					sec = sec_all_exec[i];
			}
			sec = 1.0e-3 * sec;
			#endif
			#if defined (USE_NVML)
			nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp2);
			#endif
			printf ("\t%1.3e\t%1.3e\t%1.3e\t(%zd)", sec, (th.routine_flops / sec) * 1.0e-9, (th.routine_bytes / sec) * 1.0e-9, (size_t)iter);
		}
		if (th.mode == 'c') {
			ROUTINE_TARGET(th,dim_mm,dim_nn,alpha,dev_A,lda,dev_X,beta,dev_Y);
		}
		err = cudaGetLastError();
		if (err != cudaSuccess) printf("error : %s\n", cudaGetErrorString(err));
		// -------------------------------------------

		if (th.mode == 'c') {
			cublasGetVector (th.dim_m * th.incy, th.word_size, dev_Y, 1, hst_Y_t, 1);
		}

		cudaFree (dev_A);
		cudaFree (dev_X);
		cudaFree (dev_Y);

		if (th.mode == 'c') {
			#if (defined (CHECK_CUBLAS))
			REFERENCE_TYPE *dev_A_r, *dev_X_r, *dev_Y_r;
			copy_vector (th.dim_m * th.incy, hst_Y, hst_Y_r);
			cudaMalloc ((void **) &dev_A_r, size_r * dim_mm * dim_nn);
			cudaMalloc ((void **) &dev_X_r, size_r * th.dim_n * th.incx);
			cudaMalloc ((void **) &dev_Y_r, size_r * th.dim_m * th.incy);
			cublasSetMatrix (dim_mm, dim_nn, size_r, hst_A_r, lda_cpu, dev_A_r, dim_mm);
			cublasSetVector (th.dim_n * th.incx, size_r, hst_X_r, 1, dev_X_r, 1);
			cublasSetVector (th.dim_m * th.incy, size_r, hst_Y_r, 1, dev_Y_r, 1);
			ROUTINE_REFERENCE(th,dim_mm,dim_nn,alpha_r,dev_A_r,lda,dev_X_r,beta_r,dev_Y_r);
			cublasGetVector (th.dim_m, size_r, dev_Y_r, 1, hst_Y_r, 1);
			check_result (th.dim_m * th.incy, hst_Y_t, hst_Y_r);
			cudaFree (dev_A_r);
			cudaFree (dev_X_r);
			cudaFree (dev_Y_r);
			#else
			copy_vector (th.dim_m * th.incy, hst_Y, hst_Y_r);
			ROUTINE_REFERENCE(th,dim_mm,dim_nn,alpha_r,hst_A_r,lda_cpu,hst_X_r,beta_r,hst_Y_r);
			check_result (th.dim_m * th.incy, hst_Y_t, hst_Y_r);
			#endif
		}

		#if defined (MUBLAS)
		print_mublas_info1 (&th, &mh);
		#endif
		#if defined (USE_NVML)
		printf ("\t%10u\t%10u", temp1, temp2);
		#endif
		printf ("\n");

		dim_step (&th);
	}
// --------------------------------------------

// shutdown -----------------------------------
	cudaFreeHost (hst_A);
	cudaFreeHost (hst_X);
	cudaFreeHost (hst_Y);
	if (th.mode == 'c') {
		free (hst_A_r);
		free (hst_X_r);
		free (hst_Y_r);
		cudaFreeHost (hst_Y_t);
	}

	#if defined (CUBLAS) || (defined (CHECK_CUBLAS))
	cublasDestroy (ch);
	#endif
	#if defined (MUBLAS)
	print_mublas_info2 (&th, &mh);
	mublasDestroy (&mh);
	#elif defined (CUBLAS)
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	#elif defined (CULA)
	#elif defined (ASPEN)
	ASPEN_shutdown();
	#endif
	#if defined (USE_NVML)
	nvmlShutdown();
	#endif
	cudaThreadExit();
// --------------------------------------------

#endif
	return 0;
}

