#ifndef TESTING_LEVEL2
#define TESTING_LEVEL2

// === GEMV =====================================================
#if defined (GEMV)
#define DIM2
// target routine
#if defined (MUBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) MUBLAS_GEMV(&mh,th.tranA,m,n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#elif defined (CUBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CUBLAS_GEMV(ch,cublas_trans_const_m(th.tranA),m,n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#elif defined (KBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) KBLAS_GEMV(th.tranA,m,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif defined (MAGMA)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) MAGMA_GEMV(magma_trans_const(th.tranA),m,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif defined (CULA)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CULA_GEMV(th.tranA,m,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif defined (ASPEN)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) ASPEN_GEMV(th.tranA,m,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#endif
// reference routine
#if   defined (CHECK_MPACK)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) MBLAS_GEMV(&th.tranA,m,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif defined (CHECK_CBLAS)
#if defined (PREC_S) || defined (PREC_D) || defined (PREC_W) || defined (PREC_B)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CBLAS_GEMV(CblasColMajor,cblas_trans_const_m(th.tranA),m,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#else
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CBLAS_GEMV(CblasColMajor,cblas_trans_const_m(th.tranA),m,n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#endif
#elif defined (CHECK_CUBLAS)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CUBLAS_GEMV(ch,cublas_trans_const_m(th.tranA),m,n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#endif
// %%%
#define ROUTINE_FLOPS(m,n)						(FPMADCOUNT*m*n+2*FPMULCOUNT*m+FPADDCOUNT*m)
#define ROUTINE_BYTES(m,n,s)					((2*m+n+m*n)*s) 
#define ROUTINE_MAXSIZE(gmem,margin,s)			(sqrt((gmem-margin)/s))
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)	((gmem-margin)/s/dim1)
#endif
// ==============================================================

// === SYMV =====================================================
#if defined (SYMV)
#define DIM1
// target routine
#if defined (MUBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y)
#elif defined (CUBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CUBLAS_SYMV(ch,cublas_uplo_const_m(th.uplo),n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#elif defined (KBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) KBLAS_SYMV(th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif defined (MAGMA)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) MAGMA_SYMV(magma_uplo_const(th.uplo),n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif defined (CULA)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CULA_SYMV(th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif defined (ASPEN)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) ASPEN_SYMV(th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#endif
// reference routine
#if defined (CHECK_MPACK)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) MBLAS_SYMV(&th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif defined (CHECK_CBLAS)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CBLAS_SYMV(CblasColMajor,cblas_uplo_const_m(th.uplo),n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif defined (CHECK_CUBLAS)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CUBLAS_SYMV(ch,cublas_uplo_const_m(th.uplo),n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#endif
// %%%
#define ROUTINE_FLOPS(n)						(FPMADCOUNT*n*n+2*FPMULCOUNT*n+FPADDCOUNT*n)
#define ROUTINE_BYTES(n,s)						(symv_mem(n)*s) 
#define ROUTINE_MAXSIZE(gmem,margin,s)			(sqrt((gmem-margin)/s))
#endif
// ==============================================================

// === HEMV =====================================================
#if defined (HEMV)
#define DIM1
// target routine
#if defined (MUBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y)
#elif defined (CUBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CUBLAS_HEMV(ch,cublas_uplo_const_m(th.uplo),n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#elif defined (KBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) KBLAS_HEMV(th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif defined (MAGMA)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) MAGMA_HEMV(magma_uplo_const(th.uplo),n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif defined (CULA)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CULA_HEMV(th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif defined (ASPEN)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) ASPEN_HEMV(th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#endif
// reference routine
#if defined (CHECK_MPACK)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) MBLAS_HEMV(&th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif defined (CHECK_CBLAS)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CBLAS_HEMV(CblasColMajor,cblas_uplo_const_m(th.uplo),n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#elif defined (CHECK_CUBLAS)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CUBLAS_HEMV(ch,cublas_uplo_const_m(th.uplo),n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#endif
// %%%
#define ROUTINE_FLOPS(n)						(FPMADCOUNT*n*n+2*FPMULCOUNT*n+FPADDCOUNT*n)
#define ROUTINE_BYTES(n,s)						(symv_mem(n)*s)
#define ROUTINE_MAXSIZE(gmem,margin,s)			(sqrt((gmem-margin)/s))
#endif
// ==============================================================

// === TRMV =====================================================
#if defined (TRMV)
#define DIM1
// target routine
#if defined (MUBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) MUBLAS_TRMV(&mh,th.uplo,th.tranA,th.diag,n,A,lda,Y,th.incy)
#elif defined (CUBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CUBLAS_TRMV(ch,cublas_uplo_const_m(th.uplo),cublas_trans_const_m(th.tranA),cublas_diag_const_m(th.diag),n,A,lda,Y,th.incy)
#elif defined (KBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) 
#elif defined (MAGMA)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) 
#elif defined (CULA)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CULA_TRMV(th.uplo,th.tranA,th.diag,n,A,lda,Y,th.incY)
#elif defined (ASPEN)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) 
#endif
// reference routine
#if defined (CHECK_MPACK)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) MBLAS_TRMV(&th.uplo,&th.tranA,&th.diag,n,A,lda,Y,th.incy)
#elif defined (CHECK_CBLAS)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CBLAS_TRMV(CblasColMajor,cblas_uplo_const_m(th.uplo),cblas_trans_const_m(th.tranA),cblas_diag_const_m(th.diag),n,A,lda,Y,th.incy)
#elif defined (CHECK_CUBLAS)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CUBLAS_TRMV(ch,cublas_uplo_const_m(th.uplo),cublas_trans_const_m(th.tranA),cublas_diag_const_m(th.diag),n,A,lda,Y,th.incy)
#endif
// %%%
#define ROUTINE_FLOPS(n)						(0) // !!!
#define ROUTINE_BYTES(n,s)						(trmv_mem(n)*s) 
#define ROUTINE_MAXSIZE(gmem,margin,s)			(sqrt((gmem-margin)/s))
#endif
// ==============================================================

// === TRSV =====================================================
#if defined (TRSV)
#define DIM1
// target routine
#if defined (MUBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) MUBLAS_TRSV(&mh,th.uplo,th.tranA,th.diag,n,A,lda,Y,th.incy)
#elif defined (CUBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CUBLAS_TRSV(ch,cublas_uplo_const_m(th.uplo),cublas_trans_const_m(th.tranA),cublas_diag_const_m(th.diag),n,A,lda,Y,th.incy)
#elif defined (KBLAS)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) 
#elif defined (MAGMA)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) 
#elif defined (CULA)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CULA_TRSV(th.uplo,th.tranA,th.diag,n,A,lda,Y,th.incY)
#elif defined (ASPEN)
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) 
#endif
// reference routine
#if defined (CHECK_MPACK)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) MBLAS_TRSV(&th.uplo,&th.tranA,&th.diag,n,A,lda,Y,th.incy)
#elif defined (CHECK_CBLAS)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CBLAS_TRSV(CblasColMajor,cblas_uplo_const_m(th.uplo),cblas_trans_const_m(th.tranA),cblas_diag_const_m(th.diag),n,A,lda,Y,th.incy)
#elif defined (CHECK_CUBLAS)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CUBLAS_TRSV(ch,cublas_uplo_const_m(th.uplo),cublas_trans_const_m(th.tranA),cublas_diag_const_m(th.diag),n,A,lda,Y,th.incy)
#endif
// %%%
#define ROUTINE_FLOPS(n)						(0) // !!!
#define ROUTINE_BYTES(n,s)						(trmv_mem(n)*s) 
#define ROUTINE_MAXSIZE(gmem,margin,s)			(sqrt((gmem-margin)/s))
#endif
// ==============================================================

size_t symv_mem (size_t n) {
	size_t i, v = 0;
	for (i = 0; i < n; i++) 
		v += (n - i);
	v += 3 * n;
	return v;
}

size_t trmv_mem (size_t n) {
	size_t i, v = 0;
	for (i = 0; i < n; i++) 
		v += (n - i);
	v += 2 * n;
	return v;
}

#endif
