#include <lis.h>
#include <cusparse_v2.h>

#include "testing_common.h"
#include "testing_level2s.h"
#include "testing_common.cpp"

int32_t
main (int32_t argc, char **argv)
{
	size_t i, iter = MINLOOP;
	float sec;
	GPU_TYPE alpha, beta;
	GPU_TYPE *hst_A, *hst_X, *hst_Y;
	GPU_TYPE *hst_Y_t;
	GPU_TYPE *dev_A, *dev_X, *dev_Y;
	int32_t *hst_Aind, *hst_Aptr;
	int32_t *dev_Aind, *dev_Aptr;

	REFERENCE_TYPE *hst_A_r = 0, *hst_X_r = 0, *hst_Y_r = 0, alpha_r = REFERENCE_ZERO, beta_r = REFERENCE_ZERO;
	size_t size_r = sizeof (REFERENCE_TYPE);

// testing setup ------------------------------
	testingHandle_t th;
	testingCreate (argc, argv, &th);
	dim_initialize (&th);
	print_info (&th);
// --------------------------------------------

// library setup ------------------------------
	cudaEvent_t t_start, t_stop;
	cudaEventCreate (&t_start);
	cudaEventCreate (&t_stop);
	cudaError_t err;

	#if defined (USE_NVML)
	uint32_t temp1 = 0, temp2 = 0;
	nvmlDevice_t dev;
	nvmlInit();
	nvmlDeviceGetHandleByIndex(th.devno, &dev);
	do {
		sleep(1);
		nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp1);
	} while (temp1 >= TEMPLIMIT);
	#endif

	#if defined (CUBLAS) || (defined (CHECK_CUBLAS))
	cusparseHandle_t ch;
	cusparseCreate (&ch);
	cusparseMatDescr_t descrA;
	cusparseCreateMatDescr (&descrA);
	cusparseSetMatType (descrA, CUSPARSE_MATRIX_TYPE_GENERAL);
	cusparseSetMatIndexBase (descrA, CUSPARSE_INDEX_BASE_ZERO);
	#endif
	#if defined (MUBLAS)
#ifdef TX128
	th.Tx_B_static = 128;
	th.Ty_B_static = 1;
#endif
	mublasHandle_t mh;
	mublasCreate (&mh);
	set_mublas_handler (&th, &mh);
	#elif defined (CUBLAS)
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	magma_init();
	#elif defined (CULA)
    culaInitialize();
	#elif defined (ASPEN)
	ASPEN_init(0);
	#else
	#endif
#if DEBUG_ALLSKIP == 0
// --------------------------------------------

// memory setup -------------------------------
//	matrix setup
	printf ("%%");
	LIS_MATRIX lis_A;
	LIS_VECTOR lis_x, lis_y;
	lis_initialize (&argc, &argv);
	lis_matrix_create (0, &lis_A);
	lis_vector_create (0, &lis_x);
	lis_vector_create (0, &lis_y);
	lis_matrix_set_type (lis_A, LIS_MATRIX_CSR);
	lis_input_matrix (lis_A, argv[1]);
	lis_matrix_assemble(lis_A);
	th.dim_k = lis_A->nnz;
	th.dim_n = lis_A->n;
	th.dim_m = th.dim_n;//lis_A->m;
	lis_vector_set_size (lis_x, th.dim_n, 0);
	lis_vector_set_size (lis_y, th.dim_m, 0);
// end
	size_t size_n	= sizeof (GPU_TYPE) * th.dim_n;
	size_t size_m	= sizeof (GPU_TYPE) * th.dim_m;
	size_t size_mi	= sizeof (int32_t) * (th.dim_m+1);
	size_t size_nnz	= sizeof (GPU_TYPE) * th.dim_k;
	size_t size_nnzi= sizeof (int32_t) * th.dim_k;
	cudaMallocHost ((void **) &hst_A, size_nnz);
	cudaMallocHost ((void **) &hst_Aptr, size_mi);
	cudaMallocHost ((void **) &hst_Aind, size_nnzi);
	cudaMallocHost ((void **) &hst_X, size_n);
	cudaMallocHost ((void **) &hst_Y, size_m);
	init_vector (th.dim_n, hst_X);
	init_vector (th.dim_m, hst_Y);
	init_vector (1, &alpha);
	init_vector (1, &beta);
	for (i = 0; i < th.dim_k; i++) hst_A[i] = lis_A->value[i];
	for (i = 0; i < th.dim_m+1; i++) hst_Aptr[i] = lis_A->ptr[i];
	for (i = 0; i < th.dim_k; i++) hst_Aind[i] = lis_A->index[i];
	for (i = 0; i < th.dim_n; i++) lis_x->value[i] = hst_X[i];
	for (i = 0; i < th.dim_m; i++) lis_y->value[i] = hst_Y[i];
	if (th.mode == 'c') {
		cudaMallocHost ((void **) &hst_Y_t, size_m);
		hst_A_r = (REFERENCE_TYPE *) malloc (size_nnz);
		hst_X_r = (REFERENCE_TYPE *) malloc (size_n);
		hst_Y_r = (REFERENCE_TYPE *) malloc (size_m);
		copy_vector (th.dim_k, hst_A, hst_A_r);
		copy_vector (th.dim_n, hst_X, hst_X_r);
		copy_vector (th.dim_m, hst_Y, hst_Y_r);
		copy_vector (1, &alpha, &alpha_r);
		copy_vector (1, &beta, &beta_r);
	}
// --------------------------------------------

	print_info2 (&th);

// evaluation ---------------------------------
		cudaMalloc ((void **) &dev_A, size_nnz);
		cudaMalloc ((void **) &dev_Aptr, size_mi);
		cudaMalloc ((void **) &dev_Aind, size_nnzi);
		cudaMalloc ((void **) &dev_X, size_n);
		cudaMalloc ((void **) &dev_Y, size_m);

		cublasSetVector (th.dim_k, th.word_size, hst_A, 1, dev_A, 1);
		cublasSetVector (th.dim_m+1, sizeof (int32_t), hst_Aptr, 1, dev_Aptr, 1);
		cublasSetVector (th.dim_k, sizeof (int32_t), hst_Aind, 1, dev_Aind, 1);
		cublasSetVector (th.dim_n, th.word_size, hst_X, 1, dev_X, 1);
		cublasSetVector (th.dim_m, th.word_size, hst_Y, 1, dev_Y, 1);

		#if defined (USE_NVML)
		nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp1);
		#endif

		printf ("%zd\t%zd\t--", (size_t)th.dim_m, (size_t)th.dim_n);
		get_routine_theoretial_performance (&th);

		// execution ---------------------------------
		if (th.mode == 'p') {
			#if PERF_ACCURATE
			iter = 100;//get_num_iter (&th);
			ROUTINE_TARGET(th,th.dim_m,th.dim_n,th.dim_k,alpha,descrA,dev_A,dev_Aptr,dev_Aind,dev_X,beta,dev_Y);
			// +++++ measurement +++++++
			cudaEventRecord (t_start, 0);
			for (i = 0; i < iter; i++) ROUTINE_TARGET(th,th.dim_m,th.dim_n,th.dim_k,alpha,descrA,dev_A,dev_Aptr,dev_Aind,dev_X,beta,dev_Y);
			cudaEventRecord (t_stop, 0);
			// +++++++++++++++++++++++++
			cudaEventSynchronize (t_stop);
			cudaEventElapsedTime (&sec, t_start, t_stop);
			sec = 1.0e-3 * sec / (float)iter;
			#else
			float sec_all_exec[MINLOOP];
			iter = MINLOOP;
			for (i = 0; i < iter; i++) {
				cudaEventRecord (t_start, 0);
				ROUTINE_TARGET(th,th.dim_m,th.dim_n,th.dim_k,alpha,descrA,dev_A,dev_Aptr,dev_Aind,dev_X,beta,dev_Y);
				cudaEventRecord (t_stop, 0);
				cudaEventSynchronize (t_stop);
				cudaEventElapsedTime (&sec, t_start, t_stop);
				sec_all_exec[i] = sec;
			}
			sec = 0.;
			for (i = 0; i < iter; i++) {
				if (sec_all_exec[i] > sec)
					sec = sec_all_exec[i];
			}
			sec = 1.0e-3 * sec;
			#endif
			#if defined (USE_NVML)
			nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp2);
			#endif
			printf ("\t%1.3e\t%1.3e\t%1.3e\t(%zd)", sec, (th.routine_flops / sec) * 1.0e-9, (th.routine_bytes / sec) * 1.0e-9, (size_t)iter);
		}
		if (th.mode == 'c') {
			ROUTINE_TARGET(th,th.dim_m,th.dim_n,th.dim_k,alpha,descrA,dev_A,dev_Aptr,dev_Aind,dev_X,beta,dev_Y);
		}
		err = cudaGetLastError();
		if (err != cudaSuccess) printf("error : %s\n", cudaGetErrorString(err));
		// -------------------------------------------

		if (th.mode == 'c') {
			cublasGetVector (th.dim_m, th.word_size, dev_Y, 1, hst_Y_t, 1);
		}

		cudaFree (dev_A);
		cudaFree (dev_Aptr);
		cudaFree (dev_Aind);
		cudaFree (dev_X);
		cudaFree (dev_Y);

		if (th.mode == 'c') {
			REFERENCE_TYPE *dev_A_r, *dev_X_r, *dev_Y_r;
			int32_t *dev_Aptr_r, *dev_Aind_r;
			copy_vector (th.dim_m, hst_Y, hst_Y_r);
			cudaMalloc ((void **) &dev_A_r, size_nnz);
			cudaMalloc ((void **) &dev_Aptr_r, size_mi);
			cudaMalloc ((void **) &dev_Aind_r, size_nnzi);
			cudaMalloc ((void **) &dev_X_r, size_n);
			cudaMalloc ((void **) &dev_Y_r, size_m);
			cublasSetVector (th.dim_k, size_r, hst_A_r, 1, dev_A_r, 1);
			cublasSetVector (th.dim_m+1, sizeof (int32_t), hst_Aptr, 1, dev_Aptr_r, 1);
			cublasSetVector (th.dim_k, sizeof (int32_t), hst_Aind, 1, dev_Aind_r, 1);
			cublasSetVector (th.dim_n, size_r, hst_X_r, 1, dev_X_r, 1);
			cublasSetVector (th.dim_m, size_r, hst_Y_r, 1, dev_Y_r, 1);
			ROUTINE_REFERENCE(th,th.dim_m,th.dim_n,th.dim_k,alpha_r,descrA,dev_A_r,dev_Aptr_r,dev_Aind_r,dev_X_r,beta_r,dev_Y_r);
			cublasGetVector (th.dim_m, size_r, dev_Y_r, 1, hst_Y_r, 1);
			check_result (th.dim_m, hst_Y_t, hst_Y_r);
			cudaFree (dev_A_r); 
			cudaFree (dev_Aptr_r); 
			cudaFree (dev_Aind_r); 
			cudaFree (dev_X_r);
			cudaFree (dev_Y_r);
		}

		#if defined (MUBLAS)
		print_mublas_info1 (&th, &mh);
		#endif
		#if defined (USE_NVML)
		printf ("\t%10u\t%10u", temp1, temp2);
		#endif
#if DISPINFO
		printf ("\n");
#else
#ifdef MUBLAS
		printf ("\t%zd\t", mh.Tx_B);
#else
		printf ("\t");
#endif
#endif
// --------------------------------------------

// shutdown -----------------------------------
	cudaFreeHost (hst_A);
	cudaFreeHost (hst_Aptr);
	cudaFreeHost (hst_Aind);
	cudaFreeHost (hst_X);
	cudaFreeHost (hst_Y);
	if (th.mode == 'c') {
		free (hst_A_r);
		free (hst_X_r);
		free (hst_Y_r);
		cudaFreeHost (hst_Y_t);
	}

	#if defined (CUBLAS) || (defined (CHECK_CUBLAS))
	cusparseDestroy (ch);
	#endif
	#if defined (MUBLAS)
	print_mublas_info2 (&th, &mh);
	mublasDestroy (&mh);
	#elif defined (CUBLAS)
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	#elif defined (CULA)
	#elif defined (ASPEN)
	ASPEN_shutdown();
	#endif
	#if defined (USE_NVML)
	nvmlShutdown();
	#endif

	lis_matrix_destroy (lis_A);
	lis_vector_destroy (lis_x);
	lis_vector_destroy (lis_y);
	lis_finalize ();
	cudaThreadExit();
// --------------------------------------------

#endif
	return 0;
}

