#include "testing_common.h"
#include "testing_level3.h"
#include "testing_common.cpp"

int32_t
main (int32_t argc, char **argv)
{
	size_t i, dim_mm, dim_nn, dim_kk, iter = MINLOOP;
	size_t lda, lda_cpu;
	size_t ldb, ldb_cpu;
	size_t ldc, ldc_cpu;
	size_t size_r = sizeof (REFERENCE_TYPE);
	float sec;
	GPU_TYPE alpha, beta;
	GPU_TYPE *hst_A, *hst_B, *hst_C;
	GPU_TYPE *hst_C_t;
	GPU_TYPE *dev_A, *dev_B, *dev_C;
	REFERENCE_TYPE *hst_A_r = 0, *hst_B_r = 0, *hst_C_r = 0, alpha_r = REFERENCE_ZERO, beta_r = REFERENCE_ZERO;
// testing setup ------------------------------
	testingHandle_t th;
	testingCreate (argc, argv, &th);
	dim_initialize (&th);
	print_info1 (&th);
// testing setup ------------------------------
// library setup ------------------------------
	cudaEvent_t t_start, t_stop;
	cudaEventCreate (&t_start);
	cudaEventCreate (&t_stop);
	cudaError_t err;

	#ifdef USE_NVML
	uint32_t temp1 = 0, temp2 = 0;
	nvmlDevice_t dev;
	nvmlInit();
	nvmlDeviceGetHandleByIndex(th.devno, &dev);
	do {
		sleep(1);
		nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp1);
	} while (temp1 >= TEMPLIMIT);
	#endif

	#ifdef MUBLAS 
	mublasHandle_t mh;
	mublasCreate (&mh);
	set_mublas_handler (&th, &mh);
	#elif CUBLAS
	cublasHandle_t ch;
	cublasCreate (&ch);
	cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
	if (th.cublasatomic == 1) 
		cublasSetAtomicsMode(ch, CUBLAS_ATOMICS_ALLOWED);
	else
		cublasSetAtomicsMode(ch, CUBLAS_ATOMICS_NOT_ALLOWED);
	#elif KBLAS
	#elif MAGMA
	magma_init();
	#elif CULA
    culaInitialize();
	#elif ASPEN
	ASPEN_init(0);
	#else
	#endif
// library setup ------------------------------
// memory setup -------------------------------
	cudaMallocHost ((void **) &hst_A, th.word_size * th.dim_m * th.dim_k);
	cudaMallocHost ((void **) &hst_B, th.word_size * th.dim_k * th.dim_n);
	cudaMallocHost ((void **) &hst_C, th.word_size * th.dim_m * th.dim_n);
	init_vector (1, &alpha);
	init_vector (1, &beta);
	init_vector (th.dim_m * th.dim_k, hst_A);
	init_vector (th.dim_k * th.dim_n, hst_B);
	init_vector (th.dim_m * th.dim_n, hst_C);
	if (th.mode == 'c') {
		cudaMallocHost ((void **) &hst_C_t, th.word_size * th.dim_m * th.dim_n);
		hst_A_r = (REFERENCE_TYPE *) malloc (size_r * th.dim_m * th.dim_k);
		hst_B_r = (REFERENCE_TYPE *) malloc (size_r * th.dim_k * th.dim_n);
		hst_C_r = (REFERENCE_TYPE *) malloc (size_r * th.dim_m * th.dim_n);
		copy_vector (th.dim_m * th.dim_k, hst_A, hst_A_r);
		copy_vector (th.dim_k * th.dim_n, hst_B, hst_B_r);
		copy_vector (th.dim_m * th.dim_n, hst_C, hst_C_r);
		copy_vector (1, &alpha, &alpha_r);
		copy_vector (1, &beta, &beta_r);
	}

// memory setup -------------------------------

	print_info3 (&th);

// evaluation ---------------------------------
	dim_start (&th);
	while (1) {
		if ((th.const_m == 0 && th.dim_m > th.dim_stop) ||
		    (th.const_n == 0 && th.dim_n > th.dim_stop) ||
		    (th.const_k == 0 && th.dim_k > th.dim_stop)) break;
/*
		if (th.tranA == 'T' || th.tranA == 't' || th.tranA == 'C' || th.tranA == 'c') {
			lda = th.dim_n;
			lda_cpu = th.dim_n;
			dim_mm = th.dim_n;
			dim_nn = th.dim_m;
			dim_kk = th.dim_k;
		} else {
			lda = th.dim_m;
			if (th.const_m == 1) lda_cpu = th.dim_m;
			else lda_cpu = th.dim_stop;
			dim_mm = th.dim_m;
			dim_nn = th.dim_n;
			dim_kk = th.dim_k;
		}
*/
		#if PITCHALLOC
		size_t pitch;
		cudaMallocPitch ((void **) &dev_A, &pitch, th.word_size * lda, dim_mm);
		lda = pitch/th.word_size;
		cudaMallocPitch ((void **) &dev_B, &pitch, th.word_size * ldb, dim_nn);
		ldb = pitch/th.word_size;
		cudaMallocPitch ((void **) &dev_C, &pitch, th.word_size * ldc, dim_kk);
		ldc = pitch/th.word_size;
		#else
		cudaMalloc ((void **) &dev_A, th.word_size * dim_mm * dim_kk);
		cudaMalloc ((void **) &dev_B, th.word_size * dim_kk * dim_nn);
		cudaMalloc ((void **) &dev_C, th.word_size * dim_mm * dim_nn);
		#endif

		cublasSetMatrix(dim_mm, dim_kk, th.word_size, hst_A, lda_cpu, dev_A, lda);
		cublasSetMatrix(dim_nn, dim_kk, th.word_size, hst_B, ldb_cpu, dev_B, ldb);
		cublasSetMatrix(dim_mm, dim_nn, th.word_size, hst_C, ldc_cpu, dev_C, ldc);

		#ifdef USE_NVML
		nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp1);
		#endif

		printf ("%zd\t%zd\t%zd", dim_mm, dim_nn, dim_kk);
		get_routine_theoretial_performance (&th);

		// execution ---------------------------------
		if (th.mode == 'p') {
			#if PERF_ACCURATE
			iter = get_num_iter (&th);
			ROUTINE_TARGET(th,dim_mm,dim_nn,dim_kk,alpha,dev_A,lda,dev_B,ldb,beta,dev_C,ldc); // dummy
			cudaEventRecord (t_start, 0);
			for (i = 0; i < iter; i++) ROUTINE_TARGET(th,dim_mm,dim_nn,dim_kk,alpha,dev_A,lda,dev_B,ldb,beta,dev_C,ldc);
			cudaEventRecord (t_stop, 0);
			cudaEventSynchronize (t_stop);
			cudaEventElapsedTime (&sec, t_start, t_stop);
			sec = 1.0e-3 * sec / (float)iter;
			#else
			float sec_all_exec[MINLOOP];
			iter = MINLOOP;
			for (i = 0; i < iter; i++) {
				cudaEventRecord (t_start, 0);
				ROUTINE_TARGET(th,dim_mm,dim_nn,dim_kk,alpha,dev_A,lda,dev_B,ldb,beta,dev_C,ldc);
				cudaEventRecord (t_stop, 0);
				cudaEventSynchronize (t_stop);
				cudaEventElapsedTime (&sec, t_start, t_stop);
				sec_all_exec[i] = sec;
			}
			sec = 0.;
			for (i = 0; i < iter; i++) {
				if (sec_all_exec[i] > sec) sec = sec_all_exec[i];
			}
			sec = 1.0e-3 * sec;
			#endif
		}
		if (th.mode == 'c') {
			ROUTINE_TARGET(th,dim_mm,dim_nn,dim_kk,alpha,dev_A,lda,dev_B,ldb,beta,dev_C,ldc);
		}
		// execution ---------------------------------

		err = cudaGetLastError();
		if (err != cudaSuccess) printf("error : %s\n", cudaGetErrorString(err));
		#ifdef USE_NVML
		nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp2);
		#endif

		if (th.mode == 'p') {
			printf ("\t%1.3e\t%1.3e\t%1.3e\t(%" PRId64 ")", sec, (th.routine_flops / sec) * 1.0e-9, (th.routine_bytes / sec) * 1.0e-9, iter);
		}
		if (th.mode == 'c') {
			cublasGetVector (th.dim_m * th.dim_n, th.word_size, dev_C, 1, hst_C_t, 1);
			copy_vector (th.dim_m * th.dim_n, hst_C, hst_C_r);
			ROUTINE_REFERENCE(th,dim_mm,dim_nn,dim_kk,alpha_r,hst_A_r,lda_cpu,hst_B_r,ldb_cpu,beta_r,hst_C_r,ldc_cpu);
			check_result (th.dim_m * th.dim_n, hst_C_t, hst_C_r);
		}

		#ifdef MUBLAS
		print_mublas_info1 (&th, &mh);
		#endif
		#ifdef USE_NVML
		printf ("\t%10u\t%10u", temp1, temp2);
		#endif
		printf ("\n");

		cudaFree (dev_A);
		cudaFree (dev_B);
		cudaFree (dev_C);

		dim_step (&th);
	}
// evaluation ---------------------------------
// shutdown -----------------------------------
	cudaFreeHost (hst_A);
	cudaFreeHost (hst_B);
	cudaFreeHost (hst_C);
	if (th.mode == 'c') {
		free (hst_A_r);
		free (hst_B_r);
		free (hst_C_r);
		cudaFreeHost (hst_C_t);
	}

	#ifdef MUBLAS
	print_mublas_info2 (&th, &mh);
	mublasDestroy (&mh);
	#elif CUBLAS
	cublasDestroy (ch);
	#elif KBLAS
	#elif MAGMA
	#elif CULA
	#elif ASPEN
	ASPEN_shutdown();
	#endif
	#ifdef USE_NVML
	nvmlShutdown();
	#endif
	cudaThreadExit();
// shutdown -----------------------------------

	return 0;
}

